/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"
//#include "elf_info.h"
#include "elf_pv.h"
#include <errno.h>

#define USE_MMAP 1

#if USE_MMAP
#include <sys/mman.h>
#endif /* USE_MMAP */

#ifndef DT_GNU_HASH
#define DT_GNU_HASH	0x6ffffef5	/* GNU-style hash table.  */
#endif

static bool TELF_InfoItem__Destroy(TELF_InfoItem *self) {
	xfree(self->verdef);
	xfree(self->verneed);
	xfree(self->so_name);
	return TItem__Destroy((TItem*)self);
}

static void *e_GetAdr(TELF_private *pv, Telf_offset offset, Telf_offset size) {
	if ((offset <= EI_NIDENT) || 
	    ((offset + size) > pv->e_size)) return NULL;
	return pv->e_data + offset;
}

static TELF_InfoItem *e_GetInfoItem(TELF_private *pv, const char *so_name, dword fkt) {
  TELF_Info *elf_info = pv->elf_info;
  TELF_InfoItem *ii;
	if (!elf_info || !so_name) return NULL;
	ii = elf_info->InfoItem;
	while (ii && !pccmp(ii->so_name, so_name)) ii = ii->next;
	if (!ii) {
	    ii = TItemAdd((void*)&elf_info->InfoItem, sizeof(TELF_InfoItem), 0);
	    ii->destroy = (TDESTROY)TELF_InfoItem__Destroy;
	    ii->so_name = pcdup(so_name);
	}
	ii->fkt |= fkt;
	return ii;
}

static void e_AddVerStr(char **dest, const char *s) {
	if (*dest) {
	    if (!token_find(*dest, s)) ppcadd_n(dest, 2, " ", s);
	} else {
	    *dest = pcdup(s);
	}
}

static const char *e_GetStr(TELF_private *pv, int scn_nr, Telf_offset offset) {
  Elf_Shdr *scn;
  const char *result;

	if ((scn_nr < 0) || (scn_nr >= pv->scn0_count)) return NULL;
	scn = &pv->scn0[scn_nr];
	if (scn->sh_type != SHT_STRTAB) return NULL;
	result = e_GetAdr(pv, scn->sh_offset + offset, 1);
	return result;
}

static void __attribute__ ((noinline)) verdef(TELF_private *pv, Elf_Shdr *scn) {
  Telf_offset offset;
  Telf_offset auxoffset;
  Elf_Verdef *def;
  const char *s;
  TELF_InfoItem *ii = NULL;

#if USE_ELFx64
	if ((sizeof(Elf32_Verdef)  != sizeof(Elf64_Verdef)) ||
	    (sizeof(Elf32_Verdaux) != sizeof(Elf64_Verdaux))) {
	    printf("%s(%d): TODO: 32/64 Bit \n", FL);
	    return;
	}
#endif
	offset = scn->sh_offset;
	for (int i=scn->sh_info;i>0;i--) {
	    def = e_GetAdr(pv, offset, sizeof(Elf_Verdef));
	    if (!def) return;
	    if (def->vd_version != VER_NEED_CURRENT) return;
	    auxoffset = offset + def->vd_aux;
	    for (int x=def->vd_cnt;x>0;x--) {
		Elf_Verdaux *aux;
		aux = e_GetAdr(pv, auxoffset, sizeof(Elf_Verdaux));
		if (!aux) break;
		s = e_GetStr(pv, scn->sh_link, aux->vda_name);
		if (!s) break;
		if (!ii) {
		    if (def->vd_flags == VER_FLG_BASE) ii = e_GetInfoItem(pv, s, ELF_PROVIDES);
		} else {
		    e_AddVerStr(&ii->verdef, s);
		}
		auxoffset += aux->vda_next;
	    }
	    if (!def->vd_next) return;
	    offset += def->vd_next;
	}
}

static void __attribute__ ((noinline)) verneed(TELF_private *pv, Elf_Shdr *scn) {
  Telf_offset offset;
  Telf_offset auxoffset;
  Elf_Verneed *need;
  const char *s;

#if USE_ELFx64
	if ((sizeof(Elf32_Verneed) != sizeof(Elf64_Verneed)) ||
	    (sizeof(Elf32_Vernaux) != sizeof(Elf64_Vernaux))) {
	    printf("%s(%d): TODO: 32/64 Bit \n", FL);
	    return;
	}
#endif

	offset = scn->sh_offset;
	for (int i=scn->sh_info;i>0;i--) {
	    need = e_GetAdr(pv, offset, sizeof(Elf_Verneed));
	    if (!need || (need->vn_version != VER_NEED_CURRENT)) return;

	    s = e_GetStr(pv, scn->sh_link, need->vn_file);
	    TELF_InfoItem *ii = e_GetInfoItem(pv, s, ELF_SONEEDED);
	    if (!ii) return;

	    auxoffset = offset + need->vn_aux;
	    for (int x=need->vn_cnt;x>0;x--) {
		Elf_Vernaux *aux;
		aux = e_GetAdr(pv, auxoffset, sizeof(Elf_Vernaux));
		if (!aux) break;
		s = e_GetStr(pv, scn->sh_link, aux->vna_name);
		if (!s) break;
		e_AddVerStr(&ii->verneed, s);
		auxoffset += aux->vna_next;
	    }
	    if (!need->vn_next) return;
	    offset += need->vn_next;
	}
}

static void __attribute__ ((noinline)) dynamic(TELF_private *pv, Elf_Shdr *scn) {
  Telf_offset offset;
  int mx;
  Elf_Dyn *dyn;
  const char *s;
#if USE_ELFx64
  void *pdyn;
  int size;
  Elf_Dyn mdyn;
#endif

	offset = scn->sh_offset;
	mx = scn->sh_size / scn->sh_entsize;
	for (int i=mx;i>0;i--) {
#if USE_ELFx64
	    size = sizeof(Elf32_Dyn);
	    if (pv->isElf64) size = sizeof(Elf64_Dyn);
	    pdyn = e_GetAdr(pv, offset, size);

	    if (pv->isElf64) {
		dyn = (Elf_Dyn*)pdyn;
	    } else {
		mdyn.d_tag = ((Elf32_Dyn*)pdyn)->d_tag;
		mdyn.d_un.d_ptr = ((Elf32_Dyn*)pdyn)->d_un.d_ptr;
		dyn = (Elf_Dyn*)&mdyn;
	    }
#else
	    dyn = e_GetAdr(pv, offset, sizeof(Elf_Dyn));
#endif
	    if (!dyn) break;
	    switch (dyn->d_tag) {
	    case DT_HASH:
// printf("%s(%d): DT_HASH\n", FL);
		pv->gotHASH = true;
		break;
	    case DT_GNU_HASH:
// printf("%s(%d): DT_GNU_HASH\n", FL);
		pv->gotGNUHASH = true;
		break;
	    case DT_DEBUG:
// printf("%s(%d): DT_DEBUG\n", FL);
		pv->gotDEBUG = true;
		break;

	    case DT_NEEDED:
// printf("%s(%d): DT_NEEDED\n", FL);
		s = e_GetStr(pv, scn->sh_link, dyn->d_un.d_val);
		e_GetInfoItem(pv, s, ELF_SONEEDED);
// printf("%s(%d): %s\n", FL, s);
		break;

	    case DT_SONAME:
// printf("%s(%d): DT_SONAME\n", FL);
		s = e_GetStr(pv, scn->sh_link, dyn->d_un.d_val);
		e_GetInfoItem(pv, s, ELF_SELF_NAME);
// printf("%s(%d): %s\n", FL, s);
  		if (pv->elf_info) pv->elf_info->isSO = True;

		break;

	    }
	    offset += scn->sh_entsize;
	}
}

static void elf_pv_free(TELF_private *pv) {
#if USE_MMAP
	if (pv->e_mmap) {
	    munmap(pv->e_data, pv->e_size);
	} else {
#endif /* USE_MMAP */
	    xfree(pv->e_data);
#if USE_MMAP
	}
#endif /* USE_MMAP */
	if (pv->scn0 && ((char*)pv->scn0 != pv->pscn0)) xfree(pv->scn0);
}

void elf_info_free(TELF_Info *elf_info) {
	if (elf_info) {
	    while (elf_info->InfoItem) Tfree(elf_info->InfoItem);
	    xfree(elf_info);
	}
}

TELF_Info *elf_getinfo(const char *fn, TNetErrRet *ErrRet, int fkt) {
  TELF_private pv;
  TELF_Info *elf_info;
  Telf_offset e_shentsize;
  Tstat64 *st;
  FileHandle fd;
  int errmk;

	Fill0(&pv, sizeof(pv));
	st = file_stat_fn(fn, fsFollowLink);
	if (!st) {
	    if (ErrRet) ErrRet->sys_errno = errno;
	    return NULL;
	}

	errmk = 0;
	if (st->size < EI_NIDENT) errmk = ENODATA;
	if (st->size > INT32_MAX)   errmk = EFBIG;
	if (errmk) {
	    if (ErrRet) ErrRet->sys_errno = errmk;
	    return NULL;
	}

	pv.e_size = st->size;
	xfree(st);

	elf_info = xcalloc(sizeof(TELF_Info));
	pv.elf_info = elf_info;
	if (!fkt) fkt = ELFfkt_SONEED|ELFfkt_VERNEED|ELFfkt_PROVIDES;

	fd = 0;
	if (!file_open_rd(fn, ErrRet, &fd)) goto ERR_end;

#if USE_MMAP
	void *tmp = mmap(0, pv.e_size, PROT_READ | PROT_WRITE,
				MAP_PRIVATE, fd, 0);
	if (tmp != (void*)-1) {
	    pv.e_data = tmp;
	    pv.e_mmap = True;
	} else {
#endif /* USE_MMAP */
	    pv.e_data = zmalloc(pv.e_size);
	    if (!file_read_full(fd, ErrRet, pv.e_data, 
				pv.e_size, NULL)) goto ERR_end;
#if USE_MMAP
	}
#endif /* USE_MMAP */
	if (!pv.e_data) {
	    if (ErrRet) ErrRet->err = pcdup("file too big for memory");
	    goto ERR_end;
	}

	elf_info->e_encoding = pv.e_data[EI_DATA];
	elf_info->e_osabi    = pv.e_data[EI_OSABI];

	if ((*(dword*)pv.e_data != 0x464c457f) ||
	    (pv.e_data[EI_VERSION] != EV_CURRENT) ||
#if USE_ELFx64
	    ((pv.e_data[EI_CLASS] != ELFCLASS32) &&
	     (pv.e_data[EI_CLASS] != ELFCLASS64))) goto ERR_end;
	pv.isElf64 = pv.e_data[EI_CLASS] == ELFCLASS64;
#else
	    (pv.e_data[EI_CLASS] != ELFCLASS32)) goto ERR_end;
	pv.isElf64 = False;
#endif
	elf_info->isElf64 = pv.isElf64;

	e_shentsize = 0;
#if USE_ELFx64
	if (pv.isElf64) {
	    Elf64_Ehdr *pElf_Ehdr = (Elf64_Ehdr*)pv.e_data;
	    elf_info->e_type = pElf_Ehdr->e_type;
	    elf_info->e_machine = pElf_Ehdr->e_machine;
	    pv.scn0_count = pElf_Ehdr->e_shnum;
	    if (pv.scn0_count) {
		e_shentsize = pElf_Ehdr->e_shentsize;
		if (e_shentsize < sizeof(Elf64_Shdr)) goto ERR_end;
		pv.pscn0 = e_GetAdr(&pv, pElf_Ehdr->e_shoff, pv.scn0_count * e_shentsize);
	    }
	} else {
#endif
	    Elf32_Ehdr *pElf_Ehdr = (Elf32_Ehdr*)pv.e_data;
	    elf_info->e_type = pElf_Ehdr->e_type;
	    elf_info->e_machine = pElf_Ehdr->e_machine;
	    pv.scn0_count = pElf_Ehdr->e_shnum;
	    if (pv.scn0_count) {
		e_shentsize = pElf_Ehdr->e_shentsize;
		if (e_shentsize < sizeof(Elf32_Shdr)) goto ERR_end;
		pv.pscn0 = e_GetAdr(&pv, pElf_Ehdr->e_shoff, pv.scn0_count * e_shentsize);
	    }
#if USE_ELFx64
	}
#endif

	if (pv.scn0_count && (pv.scn0_count != 0xffff) && pv.pscn0) {
#if USE_ELFx64
	    if (pv.isElf64) {
		if (e_shentsize == sizeof(Elf_Shdr)) {
		    pv.scn0 = (Elf_Shdr*)pv.pscn0;
		} else {
		    pv.scn0 = xmalloc(sizeof(Elf_Shdr) * pv.scn0_count);
		    Elf_Shdr *p64 = (Elf_Shdr*)pv.scn0;
		    char *pscn0 = pv.pscn0;
		    for (int i=pv.scn0_count;i>0;i--) {
			Elf64_Shdr *p64x = (Elf64_Shdr*)pscn0;
			*p64 = *p64x;
			p64++;
			pscn0 += e_shentsize;
		    }
		}
	    } else {
		char *pscn0 = pv.pscn0;
		pv.scn0 = xmalloc(sizeof(Elf_Shdr) * pv.scn0_count);
		Elf_Shdr *p64 = (Elf_Shdr*)pv.scn0;
		for (int i=pv.scn0_count;i>0;i--) {
		    Elf32_Shdr *p32 = (Elf32_Shdr*)pscn0;
		    p64->sh_name = p32->sh_name;
		    p64->sh_type = p32->sh_type;
		    p64->sh_flags = p32->sh_flags;
		    p64->sh_addr = p32->sh_addr;
		    p64->sh_offset = p32->sh_offset;
		    p64->sh_size = p32->sh_size;
		    p64->sh_link = p32->sh_link;
		    p64->sh_info = p32->sh_info;
		    p64->sh_addralign = p32->sh_addralign;
		    p64->sh_entsize = p32->sh_entsize;
		    p64++;
		    pscn0 += e_shentsize;
		}
	    }
#else
	    if (e_shentsize == sizeof(Elf_Shdr)) {
		pv.scn0 = (Elf_Shdr*)pv.pscn0;
	    } else {
		pv.scn0 = xmalloc(sizeof(Elf_Shdr) * pv.scn0_count);
		Elf_Shdr *p32 = (Elf_Shdr*)pv.scn0;
		char *pscn0 = pv.pscn0;
		for (int i=pv.scn0_count;i>0;i--) {
		    Elf32_Shdr *p32x = (Elf32_Shdr*)pscn0;
		    *p32 = *p32x;
		    p32++;
		    pscn0 += e_shentsize;
		}
	    }
#endif

	    Elf_Shdr *scn = pv.scn0;
	    for (int i=pv.scn0_count;i>0;i--) {
		switch (scn->sh_type) {
		case SHT_GNU_verdef:
		    if (fkt & ELFfkt_PROVIDES) verdef(&pv, scn);
		    break;

		case SHT_GNU_verneed:
		    if (fkt & ELFfkt_VERNEED) verneed(&pv, scn);
		    break;

		case SHT_DYNAMIC:
		    if (fkt & (ELFfkt_SONEED|ELFfkt_VERNEED)) dynamic(&pv, scn);
		    break;
		}
		scn++;
	    }
	}

	elf_info->gnuHASH = !pv.gotHASH && pv.gotGNUHASH;
	goto Ok_end;

ERR_end:
	elf_info_free(elf_info);
	elf_info = NULL;
Ok_end:
	elf_pv_free(&pv);
	file_close(fd, NULL);
	return elf_info;
}
