/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"
#include <unistd.h>
#include "tfio.h"

#define DEFAULT_IOfkt (IOfkt_CanWrite | IOfkt_CanInsert | IOfkt_CanSave | IOfkt_CanSaveAs)

#define TCSelf TFIOlp
typedef struct TFIOlp {
  AlgTFIO
  TLineBuf *lb;
  TglobPath *gp;
  Tstat64 *st;
  int64_t Offset;
  TNetLineBuf *nlb;
  bool IsOpen;
  bool Changed;
} TFIOlp;
#undef TCSelf


static int64_t fio_GetSizeAbs(TFIOlp *self) {
	return self->lb->datalg;
}

static int64_t TFIOlp_GetSize(TFIOlp *self) {
	return fio_GetSizeAbs(self) - self->Offset;
}

static bool fio_SetOffset(TFIOlp *self, int64_t Offset) {
  int64_t max;

	if (Offset < 0) Offset = 0;
	max = fio_GetSizeAbs(self);
	if (Offset > max) Offset = max;
	if (self->Offset != Offset) {
	    self->Offset = Offset;
	    if (self->OnSetOffset) self->OnSetOffset(self, self->Offset);
	}
	return True;
}

static int64_t fio_GetOffset(TFIOlp *self) {
	return self->Offset;
}

static char *TFIOlp_GetDataP(TFIOlp *self, int64_t Offset, int MaxChr) {
	Offset += self->Offset;
	if ((Offset < 0) || (Offset >= fio_GetSizeAbs(self))) return NULL;
	return self->lb->pc + Offset;
}

static bool TFIOlp_OverWrite(TFIOlp *self, int64_t Offset, const char *s, int size) {
  int64_t max;
  char *pc;

	if (!s || (size <= 0)) return True;
	Offset += self->Offset;
	max = fio_GetSizeAbs(self);
	if (max > 0x7fffffff) return False;
	if ((Offset < 0) || (Offset >= max)) return False;
	if ((Offset + size) >= max) size = max - Offset;
	if (size <= 0) return True;
	pc = self->lb->pc + Offset;
	pnmove(pc, s, size);
	self->Changed = True;
	return True;
}

static bool TFIOlp_Insert(TFIOlp *self, int64_t Offset, const char *s, int size) {
  int64_t max;

	if (!s || (size <= 0)) return True;
	Offset += self->Offset;
	max = fio_GetSizeAbs(self);
	if ((Offset < 0) || (max + size > 0x7fffffff)) return False;
	if ((Offset + size) > 0x7fffffff) return False;
// printf("%s(%d): %lld 0x%02x %d\n", FL, Offset, (byte)s[0], size);
	TLineBuf_Ins(self->lb, Offset, s, size);
	self->Changed = True;
	return True;
}

static bool TFIOlp_Delete(TFIOlp *self, int64_t Offset, int size) {

	if (size <= 0) return True;
	Offset += self->Offset;
	if ((Offset < 0) || (Offset > 0x7fffffff)) return False;
	TLineBuf_Del(self->lb, Offset, size);
	self->Changed = True;
	return True;
}

static void nlb_OnFileMsg(TFileMsg *fmsg) {
  TNetLineBuf *nlb;
  TFIOlp *fio;
  bool mk_eof;

// file_msg_debug(fmsg, NULL);
	if ((fmsg->art == fmArt_DataIO) && 
	    (fmsg->typD == fmTyp_ioinfo) && 
	    (fmsg->wtyp == wtyp_TNetLineBuf)) {
	    nlb = fmsg->obj;
	    fio = nlb->owner;
	    mk_eof = (fmsg->D.ioinfo->stat & fmIOInfo_EOF) != 0;
	    if (fio && fio->OnData) fio->OnData(fio, mk_eof);
	    if (mk_eof) {
// printf("%s(%d): 0x%08x\n", FL, (int)fio->nlb);
		nlb->lb = NULL;
		ppTfree(&fio->nlb);
		if (fio->OnReadEnd) fio->OnReadEnd(fio);
	    }
	}
}

static bool TFIOlp_Close(TFIOlp *self) {
	net_err_clr(&self->ErrRet);
	TLineBuf_Clr(self->lb);
	if (self->IsOpen) {
	    self->IsOpen = False;
	}
	gp_free(self->gp);
	ppfree(&self->st);
	self->Changed = False;
	self->Offset = 0;
	self->IOfkt = DEFAULT_IOfkt;
	return True;
}

static bool TFIOlp_Open(TFIOlp *self) {
  const char *pcerr = NULL;

	TFIOlp_Close(self);

	if (self->nlb) {
	    appError(EMSG_ERR1, "%s(%d): NetLineBuf not free ???", FL);
	    return False;
	}

	if (!self->FileName) goto end;

	self->gp = pctogp(self->FileName);

	gp_CryptMini16c_enc(self->gp);

	self->st = net_stat_gp(self->gp, &self->ErrRet, fsFollowLink | fsGetFileType | fsNoFileName | fsNoLinkName);
	if (!self->st || (self->st == PTR_FAILED)) {
	    if (!self->st) {
		pcerr = "file not found";
	    }
	    self->st = NULL;

	} else if (!(self->st->st & st_Reg)) {
	    pcerr = "file is not a Regular file !!!";
	}

	if (pcerr) {
	    char *fn = get_DisplayFileStr(NULL, self->gp);
	    TTC_netError((TTC0*)self, NULL, EMSG_ERR1, "%s \n%s", pcerr, fn);
	    xfree(fn);
	    goto end_msg;
	}

	self->nlb = TNetLineBuf_Create(self, self->gp);
	if (!self->nlb || !self->nlb->open) goto end_msg;
	self->nlb->ioinfo.DataSize = self->st->size;
	self->nlb->owner = self;
	self->nlb->OnFileMsg = nlb_OnFileMsg;

//	TLineBuf_Clr(self->lb);
	self->nlb->lb = self->lb;
// printf("%s(%d): \n", FL);

	if (!self->nlb->open(self->nlb, NetOpenMod_Read)) {
// printf("%s(%d): \n", FL);
	    self->nlb->lb = NULL;
	    ppTfree(&self->nlb);
	    goto end_msg;
	}
end:
//	if (self->st && (self->st->st & st_ReadOnly))
	if (self->st && !file_chk_access_st(self->st, W_OK))
	    self->IOfkt &= ~IOfkt_CanWrite;

	self->IsOpen = True;
#ifdef _OS_WIN32
 printf("%s(%d): TODO win32 lb AddPipeProc\n", FL);
	if (self->nlb->wait) self->nlb->wait(self->nlb);
#endif /* _OS_WIN32 */
end_msg:
	if (!self->IsOpen && self->OnReadEnd) self->OnReadEnd(self);
	return self->IsOpen;
}

/* static bool TFIOlp_ReOpen(TFIOlp *self, const char *FileName) {
	ppcchange_cmp(&self->FileName, FileName);
	return TFIOlp_Open(self);
} */

static void nlb_Save_OnFileMsg(TFileMsg *fmsg) {
  TNetLineBuf *nlb;
  TFIOlp *fio;

// file_msg_debug(fmsg, NULL);
	if ((fmsg->art == fmArt_DataIO) && 
	    (fmsg->typD == fmTyp_ioinfo) && 
	    (fmsg->wtyp == wtyp_TNetLineBuf)) {
	    nlb = fmsg->obj;
	    fio = nlb->owner;
	    if ((fmsg->D.ioinfo->stat & fmIOInfo_EOF) != 0) {
		nlb->lb = NULL;
		ppTfree(&fio->nlb);
		fio->Changed = False;
		if (fio->OnWriteEnd) fio->OnWriteEnd(fio);
	    }
	}
}

static bool TFIOlp_Save(TFIOlp *self, const char *FileName) {

	if (!self->IsOpen || (!FileName && !self->FileName)) return False;
	if (self->nlb) {
	    appError(EMSG_ERR1, "%s(%d): NetLineBuf not free ???", FL);
	    return False;
	}

	if (FileName) ppcchange_cmp(&self->FileName, FileName);
	if (!self->FileName) return False;

	gp_free(self->gp);
	self->gp = pctogp(self->FileName);
	gp_CryptMini16c_enc(self->gp);

	self->nlb = TNetLineBuf_Create(self, self->gp);
	if (!self->nlb) return False;

//	self->lb->lg = self->lb->datalg;
//	self->lb->pc_rdidx = NULL;
	self->nlb->OnFileMsg = nlb_Save_OnFileMsg;
	self->nlb->lb = self->lb;
	if (!self->nlb->open || !self->nlb->open(self->nlb, NetOpenMod_Write)) {
	    self->nlb->lb = NULL;
	    ppTfree(&self->nlb);
	    printf("%s(%d): ERROR: nlb->open Write\n", FL);
	    return False;
	}
	

//	self->Changed = False;
	return True;
}

/* static bool TFIOlp_SaveAs(TFIOlp *self, const char *FileName) {
	if (!self->IsOpen) return False;

	return False;
} */

static bool GetChanged(TFIOlp *self) {
	return self->Changed;
}

/*____________________________ Destroy __________________________________*/
static bool TFIOlp__Destroy(TFIOlp *self) {
	TFIOlp_Close(self);
	if (self->nlb) {
	    self->nlb->lb = NULL;
	    Tfree(self->nlb);
	}
	Tfree(self->lb);
	return TFIO__Destroy((TFIO*)self);
}

/*____________________________ Create ___________________________________*/
TFIO *TFIO_Create_LB(const char *FileName) {
  TFIOlp *self;

	self = (TFIOlp*)TFIO_Create(FileName, sizeof(TFIOlp), FIOartLineBuf);
	self->destroy = (TDESTROY)TFIOlp__Destroy;

	self->lb = TLineBuf_Create();
	self->lb->wr_memlook = True;

	self->Open = TFIOlp_Open;
//	self->ReOpen = TFIOlp_ReOpen;
	self->Save = TFIOlp_Save;
//	self->SaveAs = TFIOlp_SaveAs;
	self->GetDataP = TFIOlp_GetDataP;
	self->GetSize = TFIOlp_GetSize;
	self->GetSizeAbs = fio_GetSizeAbs;
	self->SetOffset = fio_SetOffset;
	self->GetOffset = fio_GetOffset;
	self->GetChanged = GetChanged;
	self->OverWrite = TFIOlp_OverWrite;
	self->Insert = TFIOlp_Insert;
	self->Delete = TFIOlp_Delete;

	self->IOfkt = DEFAULT_IOfkt;

	return (TFIO*)self;
}
