/*
 Copyright (C) 2010-2013 by Hoefer Alfred

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA

 ----------------------------------------------------------------
 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldftp.h"
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>

static void aftp_SendOkToSlave(TALDftp *aftp, int code);
static void aldftp_InputStop(TALDftp *aftp, bool value);

static void lim_wait(int limX, uint64_t LoadTime_us, uint64_t DatLg) {
  double mus, tus, dus;
	mus = (double)LoadTime_us;
	tus = ((double)DatLg * 976.5625) / (double)limX;
	dus = tus - mus;
	if (dus > 10000) {
	    if (dus > 900000) dus = 900000;
	    usleep(dus);
	}
}

typedef struct {
  const char *s;
  void (*CmdProc)(TALDftp *aftp, const char *s);
} TCmdBef;

static void Cmd_Close(TALDftp *aftp, const char *s) {
  bool mk;
	if (aftp->ftp->isConnect) {
	    aftp->ftp->ftpSh = ftpSh_QUIT;
	    mk = TFtp_BaseBefSend(aftp->ftp);
	    aldftp_InputStop(aftp, mk);
	}
}

static void Cmd_Cmd(TALDftp *aftp, const char *s) {
  char *toc;
  bool mk;
	if (s && *s) {
	    if (!aftp->tty) aftp->any_cmd = True;
	    GetToken(s, &toc);
	    mk = TFtp_Command(aftp->ftp, toc);
	    xfree(toc);
	    aldftp_InputStop(aftp, mk);
	}
}

static void ftp_SetURL(TFtp *ftp, const char *pc) {
	pc = pcsetgp_url(&ftp->gp, pc, 0);
	pc = pcsetgp_path(&ftp->gp, pc);
	gp_set_name(&ftp->gp, pc);
}

static const char *ftp_SetURL_Token(TFtp *ftp, const char *pc) {
  const char *result;
  char *toc;
	result = GetToken(pc, &toc);
	ftp_SetURL(ftp, toc);
	xfree(toc);
	return result;
}

static void ftpSh_BaseBef(TALDftp *aftp, int ftpSh) {
  TFtp *ftp = aftp->ftp;
	ftp->ftpSh = ftpSh;
	bool mk = TFtp_BaseBefSend(ftp);
	aldftp_InputStop(aftp, mk);
}

static void CMD_SetURL_BaseBef(TALDftp *aftp, const char *s, int ftpSh) {
  TFtp *ftp = aftp->ftp;

	if (s && *s) {
	    ftp_SetURL_Token(ftp, s);
	    ftpSh_BaseBef(aftp, ftpSh);
	}
}

static void Cmd_Connect(TALDftp *aftp, const char *s) {
	CMD_SetURL_BaseBef(aftp, s, ftpSh_Connect);
}

static void Cmd_Open(TALDftp *aftp, const char *s) {
	CMD_SetURL_BaseBef(aftp, s, ftpSh_Connect | ftpSh_USER | ftpSh_PASS);
}

static void Cmd_Ver(TALDftp *aftp, const char *s) {
  char ver[100];
  TdServHead *sh;

	snprintf(ver, sizeof(ver),
		"ALDccp: aldftp version "VERSION"  (build %d  %s)",
		ALD_BUILD_V, __DATE__);
	if (aftp->tty) {
	    printf("%s\n", ver);
	} else {
	    int lg = pclen(ver);
	    sh = xcalloc(sizeof(TdServHead) + lg);
	    sh->lg = sizeof(TdServHead)+lg;
	    sh->HeadArt = hServHead;
	    sh->bef = hSHb_VerStr;
	    sh->strg = scSHsg_MSG;
	    pcnmove(sh->dat, ver, lg);
	    write(4, sh, sh->lg);
	    xfree(sh);
	}
}

static void Cmd_Quit(TALDftp *aftp, const char *s) {
	appDat.gQUIT_Ev = True;
}

static void Ls_OnFtpDataRecv(TFtp *ftp) {
  int fuEOL;
  int tsf = 0;
  char *s;
  Tlgs lgs;
  TALDftp *aftp = ftp->owner;
  TdServFileStat *shst;

loop:
	fuEOL = fio_GetLgPlusEOL;
	if (TLineBuf_GetLineEOL_lgs(&ftp->RdDataBuf, &lgs, &fuEOL)) {
	    s = pcndup(lgs.s, lgs.lg);
// fprintf(stderr, "*:%s", s);
// fflush(stdout);
	    if (aftp->tty) {
		printf("%s", s);
	    } else {
		shst = ftptoServFile(s, &tsf);
		if (shst) {
		    write(4, shst, shst->lg);
		    xfree(shst);
		} else if (!(tsf & ftptsf_isDotFile)) {
    fprintf(stderr, "E: aftp_Ls_reply: ftptoServFile failed (\"%s\")", s);
		}
	    }
	    xfree(s);
	    goto loop;
	}
}

static void Cmd_Ls(TALDftp *aftp, const char *s) {
  TFtp *ftp = aftp->ftp;

	ppcchange(&ftp->cmd_LS, s);
	ftp->OnFtpDataRecv = Ls_OnFtpDataRecv;
	ftp->OnRecvData = NULL;
	ftpSh_BaseBef(aftp, ftpSh_TYPE_I | ftpSh_InitConn | ftpSh_LIST);
}

static void Cmd_Mlst(TALDftp *aftp, const char *s) {
  TFtp *ftp = aftp->ftp;

	TLineBuf_Clr(&ftp->RdDataBuf);
	ppcchange(&ftp->cmd_LS, s);
	ftp->OnFtpDataRecv = NULL;
	ftp->OnRecvData = NULL;
	ftpSh_BaseBef(aftp, ftpSh_TYPE_I | ftpSh_MLST);
}
static void Cmd_Mlsd(TALDftp *aftp, const char *s) {
  TFtp *ftp = aftp->ftp;

	ppcchange(&ftp->cmd_LS, s);
	ftp->OnFtpDataRecv = Ls_OnFtpDataRecv;
	ftp->OnRecvData = NULL;
	ftpSh_BaseBef(aftp, ftpSh_TYPE_I | ftpSh_InitConn | ftpSh_MLSD);
}

static void Cmd_CD(TALDftp *aftp, const char *s) {
  TFtp *ftp = aftp->ftp;

	aftp->mk_us = 0;
	aftp->sch_bef = 0;
	if (s && *s) {
	    xfree(ftp->gp.path);
	    GetToken(s, &ftp->gp.path);
	    if (!aftp->tty && ftp->AktPath) {
		char *path = path_add(ftp->gp.path, NULL);
		bool ok = pccmp(ftp->AktPath, path);
		xfree(path);
		if (ok) {
		    aftp_SendOkToSlave(aftp, 250);
		    return;
		}
	    }
	    ftpSh_BaseBef(aftp, ftpSh_CWD);
	} else {
	    fprintf(stderr, "Error: No File\n");
	}
}

/* void Cmd_Size(TALDftp *aftp, const char *s) {
  TFtp *ftp = aftp->ftp;

	ppcchange(&ftp->cmd_LS, s);
	ftpSh_BaseBef(aftp, ftpSh_SIZE);
} */

static void Cmd_Stat(TALDftp *aftp, const char *s) {
  TFtp *ftp = aftp->ftp;

	ppcchange(&ftp->cmd_LS, s);
	ftpSh_BaseBef(aftp, ftpSh_STAT);
}

static void aftp_AzRecv(TALDftp *aftp, bool last) {
  TFtp *ftp = aftp->ftp;
  uint64_t ti;
  double kbs;

	ti = GetCurTimeDiv_us64(ftp->TransfStart_us);
	if ((ti > aftp->mk_us) || last) {
	    aftp->mk_us += 250000;
	    kbs = ftp->TransfDatLg;
	    kbs = kbs * 1024;
	    kbs = kbs / ti;
#if __WORDSIZE == 64
	    fprintf(stderr, "%c %8.2fKb/s %10ldByte ", 0x0d, kbs, ftp->TransfDatLg);
#else
	    fprintf(stderr, "%c %8.2fKb/s %10lldByte ", 0x0d, kbs, ftp->TransfDatLg);
#endif
	    if (last) fprintf(stderr, "\n");
	}
}

static void Get_OnFtpDataRecv(TFtp *ftp) {
  TALDftp *aftp = ftp->owner;
	
// fprintf(stderr, "Cmd_GET \"%s\" \"%s\"\n", ftp->srcFile, ftp->destFile);
	if (ftp->destFile) {
	    if (!ftp->file_fd && !ftp->FtpStop) {
		char *fn = pcadd(ftp->destFile, ".part");
		char *dir;
		if (path_sub(fn, &dir)) {
		    ftp->FtpStop = !file_mkdir_parents_ae(dir, 0, EMSG_ERR2);
		    xfree(dir);
		}
		if (!ftp->FtpStop) {
		    ftp->FtpStop = !file_open_wr_ae(appDat.inifile_global, 
					&ftp->file_fd, 0640, EMSG_ERR2);
		}
		xfree(fn);
	    }
	    if (!ftp->FtpStop) {
		Tlgs lgs;
		if (TLineBuf_GetData_lgs(&ftp->RdDataBuf, &lgs, 0, NULL)) {
		    ftp->FtpStop = !file_write_ae(ftp->file_fd, lgs.s, lgs.lg, 
					NULL, ftp->destFile, EMSG_ERR2);
		}
	    }
	    if ((ftp->FtpStop || ftp->RdDataBuf.eof) && ftp->file_fd) {
		file_close(ftp->file_fd, NULL);
		ftp->file_fd = 0;
		char *fn = pcadd(ftp->destFile, ".part");
		if (ftp->FtpStop) {
		    unlink(fn);
		} else {
		    unlink(ftp->destFile);
		    rename(fn, ftp->destFile);
		}
		xfree(fn);
	    }
	}

	aftp_AzRecv(aftp, ftp->RdDataBuf.eof);
}

//#define TestBUFSIZ	80                

static void Get_OnRecvData(TFtp *ftp, int fd, Tfd_list *fdl, int *fu) {
  TALDftp *aftp = ftp->owner;
  int lg, blg, m;
  char line[BUFSIZ];
  TdData *Data = (TdData*)&line[0];
  char *pc;
	
	*fu = 1;
	lg = 0;
	if (aftp->tty) {
	    lg = read(fd, &line[0], BUFSIZ);
	    ftp->TransfDatLg += lg;
	    aftp_AzRecv(aftp, (lg == 0));
	} else {
// fprintf(stderr, "Get_OnRecvData: %s(%d): %d\n", FL, ftp->ftpSchr);
	    switch(ftp->ftpSchr) {
	    case ftpSchr_LIST:
		ppcfree(&ftp->srcFileStat);

		lg = read(fd, line, BUFSIZ);
		if (lg) {
		    line[lg] = 0;
		    TLineBuf_Add(&ftp->RdDataBuf, line, lg);
		    m = 0;
		    ftp->srcFileStat = ftptoServFile(line, &m);
		    write(4, ftp->srcFileStat, ftp->srcFileStat->lg);

		    lg = 0;
		    if (ftp->srcFileStat->st.st & st_Link) ftp->ftpSh |= ftpSh_SIZE;
		}
		break;

	    case ftpSchr_RETR:
		blg = BUFSIZ - sizeof(TdData);
		pc = Data->dat;
		lg = read(fd, pc, blg);
		ftp->TransfDatLg += lg;
		Data->lg = sizeof(TdData) + lg;
		Data->HeadArt = hServData;
		Data->bef = hSHbd_Dat;
		if (lg > 0) Data->fu = 0; else Data->fu = hSHfu_EOF;
		Data->sh_DataNowSize = ftp->TransfDatLg;
		Data->sh_DataSize = 0;
		if (ftp->srcFileStat) Data->sh_DataSize = ftp->srcFileStat->st.size;
// fprintf(stderr, "%s(%d): %d %d %d ges=%lld (%lld)\n", FL, sizeof(TdData), blg, lg, ftp->TransfDatLg, Data->sh_DataSize);
		Data->sh_LoadTime_us = GetCurTimeDiv_us64(ftp->TransfStart_us);

		if (aftp->limDownLoad && !Data->fu) {
		    lim_wait(aftp->limDownLoad, Data->sh_LoadTime_us, ftp->TransfDatLg);
		    Data->sh_LoadTime_us = GetCurTimeDiv_us64(ftp->TransfStart_us);
		}

		write(4, &line[0], Data->lg);
		break;
	    }
	}

	if (lg == 0) {
	    TFtp_closeconn(ftp);
	    ftp->ReplyBufAkt = False;
	    while (ReplyBuf_Line(ftp));
	}	
}

static void Cmd_Get(TALDftp *aftp, const char *s) {
  const char *ntoc;
  TFtp *ftp = aftp->ftp;

	aftp->mk_us = 0;
	aftp->sch_bef = 0;
	if (s && *s) {
	    xfree(ftp->srcFile);
	    xfree(ftp->destFile);
	    ftp->OnFtpDataRecv = NULL;
	    ftp->OnRecvData = NULL;
	    ftp->OnFtpDataSend = NULL;
	    ntoc = GetToken(s, &ftp->srcFile);
	    if (aftp->tty) {

		GetToken(ntoc, &ftp->destFile);
		if (!ftp->destFile) {
		    ftp->destFile = pcadd(xgetcwd(*ftp->srcFile != '/'), ftp->srcFile);
		}
// fprintf(stderr, "Cmd_GET \"%s\" \"%s\"\n", ftp->srcFile, ftp->destFile);

		ftp->OnFtpDataRecv = Get_OnFtpDataRecv;
	    } else {
		ppcfree(&ftp->srcFileStat);
		ftp->OnRecvData = Get_OnRecvData;
	    }
	    ftpSh_BaseBef(aftp, ftpSh_TYPE_I | ftpSh_InitConn | ftpSh_RETR);
	} else {
	    fprintf(stderr, "Error: No File\n");
	}
}

static void Cmd_UpLoad_OnFtpDataSend(TFtp *ftp) {
  TALDftp *aftp = ftp->owner;
  uint64_t m = GetCurTimeDiv_us64(ftp->TransfStart_us);
  uint64_t bufdif = ftp->TransfDatLg - ftp->TransfDatLg_Last_MSG;
  TdData shD;
  bool eof = (!ftp->WrDataBuf);

	if (!eof && aftp->limUpLoad) {
	    lim_wait(aftp->limUpLoad, m, ftp->TransfDatLg);
	    m = GetCurTimeDiv_us64(ftp->TransfStart_us);
	}

	if (eof || (m >= ftp->TransfNxEv_us) || (bufdif > 0x1e000)) {
	    ftp->TransfNxEv_us = m + 40000;
	    ftp->TransfDatLg_Last_MSG = ftp->TransfDatLg;
	    Fill0(&shD, sizeof(TdData));
	    shD.lg = sizeof(TdData);
	    shD.HeadArt = hServData;
	    shD.bef = hSHbd_Dat;
	    shD.fu = hSHfu_onlyInfo | hSHfu_send;
	    if (eof) shD.fu |= hSHfu_EOF;
	    shD.sh_DataSize = aftp->DataSize;
	    shD.sh_DataNowSize = ftp->TransfDatLg;
	    shD.sh_LoadTime_us = m;
	    write(4, &shD, shD.lg);
// fprintf(stderr, "%s(%d) OnFtpDataSend %lldus %llu %llx\n", FL, shD.sh_LoadTime_us, ftp->TransfDatLg, bufdif);
// fflush(stderr);
	}
}

static void Cmd_Put(TALDftp *aftp, const char *s) {
  const char *ntoc;
  TFtp *ftp = aftp->ftp;

	aftp->mk_us = 0;
	aftp->sch_bef = 0;
	if (s && *s) {
	    xfree(ftp->srcFile);
	    xfree(ftp->destFile);
	    ntoc = GetToken(s, &ftp->srcFile);
	    if (aftp->tty) {

		GetToken(ntoc, &ftp->destFile);
		if (!ftp->destFile) {
		    ftp->destFile = pcadd(xgetcwd(*ftp->srcFile != '/'), ftp->srcFile);
		}
 fprintf(stderr, "TODO Cmd_Put \"%s\" \"%s\"\n", ftp->srcFile, ftp->destFile);
		return;
	    } else {
// fprintf(stderr, "Cmd_Put \"%s\"\n", ftp->srcFile);
		ftp->destFile = ftp->srcFile;
		ftp->srcFile = NULL;
		ftp->OnFtpDataRecv = NULL;
		ftp->OnRecvData = Get_OnRecvData;
		ftp->OnFtpDataSend = Cmd_UpLoad_OnFtpDataSend;
	    }
	    ftpSh_BaseBef(aftp, ftpSh_TYPE_I | ftpSh_InitConn | ftpSh_STOR);
	} else {
	    fprintf(stderr, "Error: No File\n");
	}
}

static void Cmd_Help(TALDftp *aftp, const char *s);

static const TCmdBef CmdBefTab[] = {
  {"Open", Cmd_Open},
//  {"Close", Cmd_Close},
//  {"Cmd", Cmd_Cmd},
  {"Connect", Cmd_Connect},
  {"Ver", Cmd_Ver},
//  {"FileStat", Cmd_FileStat},
//  {"DownLoad", Cmd_DownLoad},
//  {"UpLoad", Cmd_UpLoad},
  {"Help", Cmd_Help},
  {"Quit", Cmd_Quit},
  {"Exit", Cmd_Quit},
  {NULL, NULL}
};

static const TCmdBef CmdOpenBefTab[] = {
  {"Cd", Cmd_CD},
  {"Close", Cmd_Close},
  {"Cmd", Cmd_Cmd},
  {"Get", Cmd_Get},
  {"Help", Cmd_Help},
  {"Ls", Cmd_Ls},
  {"Mlst", Cmd_Mlst},
  {"Mlsd", Cmd_Mlsd},
  {"Put", Cmd_Put},
//  {"Size", Cmd_Size},
  {"Status", Cmd_Stat},
  {"Quit", Cmd_Quit},
  {"Exit", Cmd_Quit},
  {NULL, NULL}
};

static void Cmd_Help(TALDftp *aftp, const char *s) {
  char line[80];
  int x, m;
  char *pc;
  const char *pcs;
  const TCmdBef *cmdb;

	cmdb = CmdBefTab;
	if (aftp->ftp->isConnect) cmdb = CmdOpenBefTab;
	m = 0;
	while (True) {
	    if (!m) {
		pc = line;
		for (x=0;x<80;x++) {*pc = 0x20; pc++;}
	    }
	    pc = &line[8+m*10];
	    pcs = cmdb->s;
	    if (pcs) {
		while (*pcs) {*pc = *pcs; pc++; pcs++;}
		m++;
		if (m > 5) {
		    *pc = '\n'; pc++; *pc=0;
		    printf(line);
		    m = 0;
		}
	    } else {
		if (m) {
		    *pc = '\n'; pc++; *pc=0;
		    printf(line);
		}
		return;
	    }
	    cmdb++;
	}
}

static void DecodeCmd(TALDftp *aftp, const char *line) {
  const char *toc, *cpc;
  int tlg;
  const TCmdBef *cmdb;

	if (!line || !*line) return;
	aftp->any_cmd = False;
	ppcfree(&aftp->ftp->lastErrStr);

	cpc = pcGetToken(line, &toc, &tlg);

	cmdb = CmdBefTab;
	if (aftp->ftp->isConnect) cmdb = CmdOpenBefTab;

	while (cmdb->s) {
	    if ((tlg == pclen(cmdb->s)) && !pcncasecmpASCII(cmdb->s, toc, tlg)) {
		cmdb->CmdProc(aftp, cpc);
		return;
	    }
	    cmdb++;
	}

// fprintf(stderr, "Error: aftp: DecodeCmd: %d %s\n", lg, &line[0]);
	TFtp_Error(aftp->ftp, ioErr_DecodeCmd, "aftp: DecodeCmd \"%s\"\n", line);
}

#ifdef USE_AH_TTY
static void tty_OnReturn(Ttty *tty, char *line, int lg) {
	DecodeCmd(tty->owner, line);
}
#endif

static void DoFtpBefEnd(TALDftp *aftp, int sch) {
// printf("> DoFtpBefEnd:\n");
#ifdef USE_AH_TTY
	if (aftp->tty) {
	    aldftp_InputStop(aftp, False);
	}
#endif
}

static void StdIn(int fd, Tfd_list *fdl) {
  TALDftp *aftp = fdl->obj;
  int lg;
	lg = TLineBuf_fd_Add(&aftp->noTTY_InBuf, 0);
	if (!lg) appDat.gQUIT_Ev = True;
}

static void aftp_Pipe3_DataIn(int fd, Tfd_list *fdl) {
  TALDftp *aftp = fdl->obj;
  int lg, dlg;
  int tsf = 0;
  TdServHead *sh;
  char line[BUFSIZ];

	lg = read(fd, line, BUFSIZ);
	if (lg <= 0) {
// fprintf(stderr, "%s(%d) OnDataPipeRead Close\n", FL);
	    ClosePipeProc(fd, 0);
	    return;
	}

	TLineBuf_Add(&aftp->dataInBuf, line, lg);
	while (GetSyncPipeServHead(&aftp->dataInBuf, &sh, &tsf)) {
	    if (sh->HeadArt == hServData) {
		dlg = sh->lg - sizeof(TdData);
		if (aftp->ftp->data && aftp->ftp->WrDataBuf) {
		    aftp->DataSize = ((TdData*)sh)->sh_DataSize;
		    TLineBuf_Add(aftp->ftp->WrDataBuf, ((TdData*)sh)->dat, dlg);
		    if (sh->fu && hSHfu_EOF) aftp->ftp->WrDataBuf->eof = True;
		}
	    } else {
 fprintf(stderr, "%s(%d) ftp_pcl_OnDataPipeRead: sh->HeadArt=%d bef=%d\n", FL, sh->HeadArt, sh->bef);
	    }
	}
}

static void aftp_DoError(TFtp *ftp, int ErrNr, const char *msg) {

	ppcchange(&ftp->lastErrStr, msg);
	ftp->lastErrCode = ErrNr;
// fprintf(stderr, "DoError %d %s\n", ErrNr, msg);
// fflush(stderr);

	if ((ErrNr >= ioErr_Unknown_host) && (ErrNr < ioErr_NoConnMax)) {
	    int lg = pclen(msg);
	    TdServHead *sh;
	    sh = xcalloc(sizeof(TdServHead) + lg);
	    sh->lg = sizeof(TdServHead) + lg;
	    sh->HeadArt = hServHead;
	    sh->bef = hSHb_ERR;
	    sh->strg = scSHsg_MSG;
	    sh->m = ErrNr;
	    pcnmove(sh->dat, msg, lg);
	    write(4, sh, sh->lg);
	    xfree(sh);
	}
}


static void aftp_DoStrgMSG(TFtp *ftp, int strg) {
  TdServHead sh;
//	if (!aftp->tty) {
// fprintf(stderr, "%s(%d) aftp_DoStrgMSG:\n", __FILE__, __LINE__);
	    sh.lg = sizeof(TdServHead);
	    sh.HeadArt = hServHead;
	    sh.bef = hSHb_StrgMSG;
	    sh.strg = strg;
	    write(4, &sh, sh.lg);
//	}
}

static void aftp_SendOkToSlave(TALDftp *aftp, int code) {
  TFtp *ftp;

	if (!aftp->tty) {
// fprintf(stderr, "SendOkToSlave %d\n", code);
// fflush(stderr);
	    ftp = aftp->ftp;
	    if (ftp->lastErrStr) {
		TdServHead *sh;
		int lg = pclen(ftp->lastErrStr);
		sh = xcalloc(sizeof(TdServHead) + lg);
		sh->lg = sizeof(TdServHead) + lg;
		sh->HeadArt = hServHead;
		if (ftp->lastErrCode == ioErr_NoLogin) {
		    sh->bef = hSHb_SET_Login;
		} else {
		    sh->bef = hSHb_ERR;
		}
		sh->m = ftp->lastErrCode;
		if (lg > 0) {
		    sh->strg = scSHsg_MSG;
		    pcnmove(sh->dat, ftp->lastErrStr, lg);
		}
		write(4, sh, sh->lg);
		xfree(sh);
		ppcfree(&ftp->lastErrStr);

	    } else {
		TdServHead sh;
		Fill0(&sh, sizeof(TdServHead));
		sh.lg = sizeof(TdServHead);
		sh.HeadArt = hServHead;
		sh.bef = hSHb_Ok;
		sh.m = code;
		write(4, &sh, sh.lg);
	    }
	}
}

static void FtpOnLineRecv(TFtp *ftp, int code, char *msg, bool last) {
  TALDftp *aftp = ftp->owner;

	if (last) printf("> %s\n", msg); else printf(">.%s\n", msg);
	fflush(stdout);

// fprintf(stderr, "Recv:%d %d %d %s\n", ftp->ftpSchr, last, code, msg);
// fflush(stderr);

	if (aftp->any_cmd) {
	    aftp->any_cmd = False;
	    if (code >= 300) {
		ppcchange(&ftp->lastErrStr, msg);
		ftp->lastErrCode = code;
	    }
	    aftp_SendOkToSlave(aftp, code);

	} else if (!aftp->tty && !last && 
		   (ftp->ftpSchr == ftpSchr_STAT) &&
		   ((code == 212) || (code == 213))) {
// fprintf(stderr, "Recv:%d %d %d %s\n", ftp->ftpSchr, last, code, msg);
	    int tsf = 0;
	    TdServFile *dsf;
	    dsf = ftptoServFile(msg, &tsf);
	    if (dsf) {
		write(4, dsf, dsf->lg);
		xfree(dsf);
	    }

	} else if ((code == 221) || (code == 421)) {
	    aftp->sch_open = 0;
	    if (last) {
		TFtp_Close(ftp);
		if (aftp->DoFtpBefEnd) aftp->DoFtpBefEnd(aftp, 0);

// fprintf(stderr, "aftp A FtpOnLineRecv: ftpSh_Err=%d\n", ftp->ftpSh_Err);
		aftp_SendOkToSlave(aftp, code);
	    }
	} else if ((code == 0) || !last) {

	} else if (TFtp_BaseBefRecv(ftp, code, msg)) {
	    if (ftp->ftpSh == 0) {
//		aldftp_InputStop(aftp, False);
		if (aftp->DoFtpBefEnd) aftp->DoFtpBefEnd(aftp, ahsLs_end);
// fprintf(stderr, "aftp B FtpOnLineRecv: ftpSh_Err=%d\n", ftp->ftpSh_Err);
		aftp_SendOkToSlave(aftp, code);
	    }
	} else {
	    if (aftp->DoFtpBefEnd) aftp->DoFtpBefEnd(aftp, 0);
	}
}

static void aldftp_InputStop(TALDftp *aftp, bool value) {
#ifdef USE_AH_TTY
	if (aftp->tty) {
	    aftp->tty->InputStop = value;
	    if (!value) Ttty_NewLine(aftp->tty, NULL, 0);
	}
#endif
}

static void ftp_STOP_closeconn(TALDftp *aftp) {
  TFtp *ftp = aftp->ftp;
	if (ftp->data <= 0) return;
	
	TFtp_closeconn(ftp);
	ftp->ReplyBufAkt = False;
	while (ReplyBuf_Line(ftp));
}


static const char *Help_Msg = {
  "\n"
  "Usage: aldftp [options]\n"
  "or     aldftp [options] user:passwd@URL\n"
  "\n"
  "  -ld                  limit download KB/sec\n"
  "  -lu                  limit upload   KB/sec\n"
  "  -h, --help           Prints out this screen\n"
  "  ALDccp: aldftp version "VERSION"  (build %d  %s)\n"
};

//int build_v = ALD_BUILD_V;
//char *build_date = __DATE__;

int main (int argc, char **argv) {
  int i;
  int limDownLoad;
  int limUpLoad;
  char *val, *open_argv, *pc;
  bool help = False;
  bool err = False;

  TALDftp aftp;

	cmdline_slash_korr(argc, argv);
	limDownLoad = limUpLoad = 0;
	open_argv = NULL;
	i = 1;
	while (!err && !help && (i<argc)) {
	    val = argv[i];
	    if (*val == '-') {
		val++;
		switch (*val) {
		case 'h':
		    help = True;
		    break;

		case 'l':
		    if (val[1] == 'd') {
			err = cmd_get_int(argc, argv, val+1, &i, &limDownLoad);
		    } else if (val[1] == 'u') {
			err = cmd_get_int(argc, argv, val+1, &i, &limUpLoad);
		    } else {
			err = cmd_get_int(argc, argv, val, &i, &limDownLoad);
			limUpLoad = limDownLoad;
		    }
		    break;

		case '-':
		    val++;
		    if (pcncmp(val, "help", 4)) {
			help = True;
		    } else err = True;
		    break;
		default: 
		    err = True;
		}
	    } else {
		if (open_argv) err = True;
		open_argv = val;
		break;
	    }
	    i++;
	}
	if (help || err) {
	    printf(Help_Msg, ALD_BUILD_V, __DATE__);
	    exit(err);
	}

	initapp(0, argc, argv);
	Fill0(&aftp, sizeof(TALDftp));
	aftp.wtyp = wtyp_USER;

#ifdef USE_AH_TTY
  Ttty *tty;
	tty = Ttty_Create(0, 1, "aldftp");
	aftp.tty = tty;
	aftp.ftp = TFtp_Create();
	if (tty) {
//		tty_is_on = True;
	    tty->owner = &aftp;
	    tty->OnKey = Ttty_LineKey;
	    tty->OnReturn = tty_OnReturn;
	} else {
	    AddPipeProc(0, InPipe, StdIn, &aftp);
	    AddPipeProc(3, InPipe, aftp_Pipe3_DataIn, &aftp);
	    aftp.ftp->DoError = aftp_DoError;
	    aftp.ftp->DoStrgMSG = aftp_DoStrgMSG;
	}
#else
	fdl = AddPipeProc(0, InPipe, StdIn);
	aftp.ftp = TFtp_Create();
	fdl->obj = &aftp;
	fdl = AddPipeProc(3, InPipe, aftp_Pipe3_DataIn);
	fdl->obj = &aftp;
#endif
	aftp.limDownLoad = limDownLoad;
	aftp.limUpLoad   = limUpLoad;

	aftp.AzTransf = False;
	aftp.ReplyBufAkt = False;
	aftp.ftp->OnLineRecv = FtpOnLineRecv;
	aftp.DoFtpBefEnd = DoFtpBefEnd;
	aftp.ftp->owner = &aftp;


	if (open_argv) {
	    pc = pcadd("open ", open_argv);
	    printf("%s\n", pc);
	    DecodeCmd(&aftp, pc);
	    xfree(pc);
	}

	if (tty) {
	    while (AppHandleMessage());
	} else {
	    while (AppHandleMessage()) {
		Tlgs lgs;
		if (TLineBuf_GetLineEOL_lgs(&aftp.noTTY_InBuf, &lgs, NULL)) {
		    char *cmd = pcndup(lgs.s, lgs.lg);
// fprintf(stderr, "lb: %d %s\n", lgs.lg, cmd);
		    if (!pccasecmpASCII(cmd, "stop")) {
			ftp_STOP_closeconn(&aftp);
		    } else {
			DecodeCmd(&aftp, cmd);
		    }
		    xfree(cmd);
		}
		AppSetWait1_single();
	    }
	}
#ifdef USE_AH_TTY
	Tfree(tty);
#endif
// fprintf(stderr, "%s(%d) aftp Exit\n", FL);
// fflush(stderr);

	return 0;
}
